#include <algorithm>
#include "uservars.h"
#include "epgsearchcats.h"
#include "epgsearchtools.h"
#include "log.h"

cUserVars UserVars;


// cUserVar
string cUserVar::Evaluate(const cEvent* e)
{
    if (oldEvent && oldEvent == e)
	return oldResult;
    usedVars.clear();
    string result;
    if (IsShellCmd())
	result = EvaluateShellCmd(e);
    else if (IsCondExpr())
	result = EvaluateCondExpr(e);
    else
	result = EvaluateCompExpr(e);
    oldResult = result;
    oldEvent = e;

    // avoid double dir separators
    int pos = 0;
    while((pos = result.find("~~")) >= 0)
	result.replace(pos, 2, "~"); 

    return result;
}

string cUserVar::EvaluateShellCmd(const cEvent* e)
{
    if (!varparser.cmd) return "";
    string cmdArgs;
    for(unsigned int i=0; i<varparser.cmdArgs.size(); i++)
    {
	string arg = varparser.cmdArgs[i];
	varparser.compExpr = arg; //handle the args as composed expr
	arg = EscapeString(EvaluateCompExpr(e));
	cmdArgs += " '" + arg + "'";
    }
    const char* res = varparser.cmd->Execute(cmdArgs.c_str());
    string result = res?res:"";
    int crPos = 0; // remove any CR
    while((crPos = result.find("\n")) >= 0)
	result.replace(crPos, 1, ""); 
    
    return result;
}

string cUserVar::EvaluateCondExpr(const cEvent* e)
{
    string condresult = "";
    cVarExpr leftExp(varparser.condEqLeft);
    string resLeft = leftExp.Evaluate(e);
    cVarExpr rightExp(varparser.condEqRight);
    string resRight = rightExp.Evaluate(e);
    if (varparser.condOp == condEq && resLeft == resRight) condresult = "1"; // assign any value
    if (varparser.condOp == condNeq && resLeft != resRight) condresult = "1"; // assign any value
    
    cUserVar* condVarTrue = UserVars.GetFromName(varparser.condvarTrue);
    cUserVar* condVarFalse = UserVars.GetFromName(varparser.condvarFalse);
    if (!condVarTrue || !condVarFalse) return "";

    if (!AddDepVar(condVarTrue)) return "";
    if (!AddDepVar(condVarFalse)) return "";

    if (condresult != "")
	return condVarTrue->Evaluate(e);
    else
	return condVarFalse->Evaluate(e);
}

string cUserVar::EvaluateCompExpr(const cEvent* e)
{
    string expr = varparser.compExpr;
    string compExprLower = expr;
    std::transform(compExprLower.begin(), compExprLower.end(), compExprLower.begin(), tolower);

    // handle internal vars like title and subtitle
    expr = EvaluateInternalVars(expr, e);

    // replace ext. EPG variables
    expr = EvaluateExtEPGVars(expr, e);

    // replace other user vars
    expr = EvaluateUserVars(expr, e);

    return expr;
}

string cUserVar::EvaluateInternalVars(const string& Expr, const cEvent* e)
{
    string expr = Expr;

    std::map<string, cInternalVar*>::iterator it;
    for (it = UserVars.internalVars.begin(); it != UserVars.internalVars.end(); it++) 
    {
	string varName = (it->second)->Name();
	int varPos = 0;
	while((varPos = FindIgnoreCase(expr, varName)) >= 0)
	{
	    usedVars.insert(it->second); 		
	    expr.replace(varPos, varName.size(), (it->second)->Evaluate(e));
	}
    }
    return expr;
}

string cUserVar::EvaluateExtEPGVars(const string& Expr, const cEvent* e)
{
    string expr = Expr;

    std::map<string, cExtEPGVar*>::iterator evar;
    for (evar = UserVars.extEPGVars.begin(); evar != UserVars.extEPGVars.end(); evar++) 
    {
	string varName = evar->second->Name();
	int varPos = 0;
	while((varPos = FindIgnoreCase(expr, varName)) >= 0)
	{
	    expr.replace(varPos, varName.size(), evar->second->Evaluate(e));
	    usedVars.insert(evar->second); 		
	}
    }
    return expr;
}

string cUserVar::EvaluateUserVars(const string& Expr, const cEvent* e)
{
    string expr = Expr;

    std::set<cUserVar*>::iterator it;
    for (it = UserVars.userVars.begin(); it != UserVars.userVars.end(); it++) 
    {
	string varName = (*it)->Name();
	int varPos = 0;
	while((varPos = FindIgnoreCase(expr, varName)) >= 0)
	{
	    if (!AddDepVar(*it)) return "";
	    expr.replace(varPos, varName.size(), (*it)->Evaluate(e));
	}
    }
    return expr;
}

string cUserVar::EvaluateInternalTimerVars(const string& Expr, const cTimer* t)
{
    string expr = Expr;

    std::map<string, cTimerVar*>::iterator tvar;
    for (tvar = UserVars.internalTimerVars.begin(); tvar != UserVars.internalTimerVars.end(); tvar++) 
    {
	string varName = tvar->second->Name();
	int varPos = 0;
	while((varPos = FindIgnoreCase(expr, varName)) >= 0)
	{
	    expr.replace(varPos, varName.size(), tvar->second->Evaluate(t));
	}
    }
    return expr;
}

bool cUserVar::DependsOnVar(const string& varName)
{
    string VarName = Strip(varName);
    cUserVar* var = UserVars.GetFromName(VarName);
    if (!var) return false;
    return DependsOnVar(var);
}

bool cUserVar::DependsOnVar(cUserVar* var)
{
    if (!var) return false;
    if (usedVars.find(var) != usedVars.end()) return true;
    std::set<cUserVar*>::iterator it;
    for (it = usedVars.begin(); it != usedVars.end(); it++) 
	if ((*it)->DependsOnVar(var))
	    return true;
    return false;
}

bool cUserVar::AddDepVar(cUserVar* var)
{
    if (var == this || var->DependsOnVar(this))
    {
	LogFile.eSysLog("ERROR - found cylic reference to var '%s' in var '%s'", var->Name().c_str(), Name().c_str());
	return false;
    }
    usedVars.insert(var); 		    
    return true;
}

void cUserVar::ResetCache()
{
    oldEvent = NULL;
    oldResult = "";
}

// cUserVarLine
bool cUserVarLine::Parse(char *s)
{
    if (s && s[0] == '#')
	return true;
    char *p = strchr(s, '=');
    if (p) 
    {
	cUserVar* userVar = new cUserVar;
	if (userVar->varparser.Parse(s))
	{
	    UserVars.userVars.insert(userVar);
	    return true;
	}
    }
    return false;
}

// cUserVars
cUserVar* cUserVars::GetFromName(const string& varName)
{
    string VarName = Strip(varName);
    std::transform(VarName.begin(), VarName.end(), VarName.begin(), tolower);
    
    std::map<string, cInternalVar*>::iterator ivar = internalVars.find(VarName);
    if (ivar != internalVars.end()) 
	return ivar->second;

    std::set<cUserVar*>::iterator uvar;
    for (uvar = userVars.begin(); uvar != userVars.end(); uvar++) 
	if (EqualsNoCase((*uvar)->Name(), VarName))
	    return (*uvar);

    std::map<string, cExtEPGVar*>::iterator evar = extEPGVars.find(VarName);
    if (evar != extEPGVars.end()) 
	return evar->second;

    LogFile.eSysLog("var '%s' not defined!", VarName.c_str());
    return NULL;
}

// cVarExpr
string cVarExpr::Evaluate(const cEvent* e)
{
    // create a dummy user var
    cUserVar var;
    if (!var.varparser.ParseExp(expr))
	return "";
    string result = var.Evaluate(e);
    usedVars = var.usedVars;
    return result;
}

string cVarExpr::Evaluate(const cTimer* t)
{
    // create a dummy user var
    cUserVar var;
    if (!var.varparser.ParseExp(expr))
	return "";

    usedVars = var.usedVars;
    return var.EvaluateInternalTimerVars(expr, t);
}

bool cVarExpr::DependsOnVar(const string& varName, const cEvent* e)
{
    string VarName = Strip(varName);
    if (FindIgnoreCase(expr, VarName) >= 0)
	return true;
    // create a dummy user var
    cUserVar var;
    var.varparser.varName = expr;
    if (!var.varparser.ParseExp(expr))
	return false;
    var.Evaluate(e);
    return var.DependsOnVar(VarName);
}
