#ifndef __USERVARS_INC__
#define __USERVARS_INC__

#include <algorithm>
#include <string>
#include <set>
#include <map>
#include <sstream>
#include <vdr/plugin.h>
#include "varparser.h"
#include "epgsearchtools.h"
#include "epgsearchcats.h"

using std::string;
using std::set;
using std::map;
using std::ostringstream;

string IntToString(int i);
int FindIgnoreCase(const string& expr, const string& query);

class cUserVar : public cListObject {
    const cEvent* oldEvent; // cache 
    string oldResult;
    string EvaluateCondExpr(const cEvent* e);
    string EvaluateCompExpr(const cEvent* e);
    string EvaluateShellCmd(const cEvent* e);
public:
    cVarParser varparser;
    set<cUserVar*> usedVars;
    virtual string Evaluate(const cEvent* e);

    string EvaluateInternalVars(const string& Expr, const cEvent* e);
    string EvaluateInternalTimerVars(const string& Expr, const cTimer* t);
    string EvaluateExtEPGVars(const string& Expr, const cEvent* e);
    string EvaluateUserVars(const string& Expr, const cEvent* e);
    virtual string Name() { return varparser.varName; }
    virtual bool IsCondExpr() { return varparser.IsCondExpr(); }
    virtual bool IsShellCmd() { return varparser.IsShellCmd(); }
    bool DependsOnVar(const string& varName);
    bool DependsOnVar(cUserVar* var);
    bool AddDepVar(cUserVar* var);
    void ResetCache();
};

class cExtEPGVar : public cUserVar {
 public:
    const string name;
    cExtEPGVar(const string& Name) : name(Name) {}
    string Name() { return name; }
    bool IsCondExpr() { return false; }
    bool IsShellCmd() { return false; }
    string Evaluate(const cEvent* e) 
	{ 
	    if (!e) return "";

	    cSearchExtCat* SearchExtCat = SearchExtCats.First();
	    while (SearchExtCat) 
	    {
		string varName = string("%") + SearchExtCat->name + string("%");
		int varPos = FindIgnoreCase(varName, name);
		if (varPos == 0)
		{
		    char* value = GetExtEPGValue(e, SearchExtCat);
		    return value?value:"";
		}
		SearchExtCat = SearchExtCats.Next(SearchExtCat);
	    }
	    return "";
	}
};

class cInternalVar : public cUserVar {
    const string name;
 public:
    cInternalVar(const string& Name) : name(Name) {}
    string Name() { return name; }
    bool IsCondExpr() { return false; }
    bool IsShellCmd() { return false; }
};

class cTitleVar : public cInternalVar {
public:
    cTitleVar() : cInternalVar("%title%") {}
    string Evaluate(const cEvent* e) { return (e && !isempty(e->Title()))? e->Title() : ""; }
};

class cSubtitleVar : public cInternalVar {
public:
    cSubtitleVar() : cInternalVar("%subtitle%") {}
    string Evaluate(const cEvent* e) { return (e && !isempty(e->ShortText()))? e->ShortText() : ""; }
};

class cSummaryVar : public cInternalVar {
public:
    cSummaryVar() : cInternalVar("%summary%") {}
    string Evaluate(const cEvent* e) { return (e && !isempty(e->Description()))? e->Description() : ""; }
};

class cHTMLSummaryVar : public cInternalVar {
public:
    cHTMLSummaryVar() : cInternalVar("%htmlsummary%") {}
    string Evaluate(const cEvent* e) 
	{ 
	    if (e && !isempty(e->Description()))
	    {
		string tmp = e->Description();
		return ReplaceAll(e->Description(), "\n", "<br />");
	    }
	    else
		return "";
	}
};

class cEventIDVar : public cInternalVar {
public:
    cEventIDVar() : cInternalVar("%eventid%") {}
    string Evaluate(const cEvent* e) { if (e) return IntToString(e->EventID()); else return ""; }
};

class cTimeVar : public cInternalVar {
public:
    cTimeVar() : cInternalVar("%time%") {}
    string Evaluate(const cEvent* e) { return (e? *(e->GetTimeString()) : ""); }
};

class cTimeEndVar : public cInternalVar {
public:
    cTimeEndVar() : cInternalVar("%timeend%") {}
    string Evaluate(const cEvent* e) { return (e? *(e->GetEndTimeString()) : ""); }
};

class cTime_wVar : public cInternalVar {
public:
    cTime_wVar() : cInternalVar("%time_w%") {}
    string Evaluate(const cEvent* e) { return (e? WEEKDAYNAME(e->StartTime()) : ""); }
};

class cTime_dVar : public cInternalVar {
public:
    cTime_dVar() : cInternalVar("%time_d%") {}
    string Evaluate(const cEvent* e) 
	{ 
	    if (!e) return "";
	    char day[3] = "";
	    struct tm tm_r;
	    const time_t t = e->StartTime();
	    tm *tm = localtime_r(&t, &tm_r);
	    strftime(day, sizeof(day), "%d", tm);
	    return day;
	}
};

class cTime_lngVar : public cInternalVar {
public:
    cTime_lngVar() : cInternalVar("%time_lng%") {}
    string Evaluate(const cEvent* e) 
	{ 
	    if (!e) return "";
	    ostringstream os;
	    os << e->StartTime();
	    return os.str();
	}
};

class cDateVar : public cInternalVar {
public:
    cDateVar() : cInternalVar("%date%") {}
    string Evaluate(const cEvent* e) 
	{ 
	    if (!e) return "";
	    char date[9] = "";
	    struct tm tm_r;
	    const time_t t = e->StartTime();
	    tm *tm = localtime_r(&t, &tm_r);
	    strftime(date, sizeof(date), "%d.%m.%y", tm);
	    return date;
	}
};

class cDateShortVar : public cInternalVar {
public:
    cDateShortVar() : cInternalVar("%datesh%") {}
    string Evaluate(const cEvent* e) 
	{ 
	    if (!e) return "";
	    char dateshort[7] = "";
	    struct tm tm_r;
	    const time_t t = e->StartTime();
	    tm *tm = localtime_r(&t, &tm_r);
	    strftime(dateshort, sizeof(dateshort), "%d.%m.", tm);
	    return dateshort;
	}
};

class cChannelNrVar : public cInternalVar {
public:
    cChannelNrVar() : cInternalVar("%chnr%") {}
    string Evaluate(const cEvent* e) 
	{ 
	    int chnr = ChannelNrFromEvent(e);
	    if (chnr < 0) return "";
	    return IntToString(chnr);
	}
};

class cChannelShortVar : public cInternalVar {
public:
    cChannelShortVar() : cInternalVar("%chsh%") {}
    string Evaluate(const cEvent* e) 
	{ 
	    if (!e) return "";
	    cChannel *channel = Channels.GetByChannelID(e->ChannelID(), true);
	    return channel?channel->ShortName(true):"";
	}
};

class cChannelLongVar : public cInternalVar {
public:
    cChannelLongVar() : cInternalVar("%chlng%") {}
    string Evaluate(const cEvent* e) 
	{ 
	    if (!e) return "";
	    cChannel *channel = Channels.GetByChannelID(e->ChannelID(), true);
	    return channel?channel->Name():"";
	}
};

class cChannelDataVar : public cInternalVar {
 public:
    cChannelDataVar() : cInternalVar("%chdata%") {}
    string Evaluate(const cEvent* e) 
	{ 
	    if (!e) return "";
	    cChannel *channel = Channels.GetByChannelID(e->ChannelID(), true);
	    return channel?CHANNELSTRING(channel):"";
	}
};

// independet variables
class cColonVar : public cInternalVar {
 public:
    cColonVar() : cInternalVar("%colon%") {}
    string Evaluate(const cEvent*) 
	{ 
	    return ":";
	}
};
 
class cDateNowVar : public cInternalVar {
 public:
    cDateNowVar() : cInternalVar("%datenow%") {}
    string Evaluate(const cEvent*) 
	{ 
	    char date[9] = "";
	    struct tm tm_r;
	    const time_t t = time(NULL);
	    tm *tm = localtime_r(&t, &tm_r);
	    strftime(date, sizeof(date), "%d.%m.%y", tm);
	    return date;
	}
};

class cDateShortNowVar : public cInternalVar {
public:
    cDateShortNowVar() : cInternalVar("%dateshnow%") {}
    string Evaluate(const cEvent*) 
	{ 
	    char dateshort[7] = "";
	    struct tm tm_r;
	    const time_t t = time(NULL);
	    tm *tm = localtime_r(&t, &tm_r);
	    strftime(dateshort, sizeof(dateshort), "%d.%m.", tm);
	    return dateshort;
	}
};

class cTimeNowVar : public cInternalVar {
public:
    cTimeNowVar() : cInternalVar("%timenow%") {}
    string Evaluate(const cEvent*) 
	{ 
	    return TIMESTRING(time(NULL));
	}
};

// timer variables
class cTimerVar {
    const string name;
 public:    
    cTimerVar(const string& Name) : name(Name) {}
    virtual ~cTimerVar() {}
    string Name() { return name; }
    virtual string Evaluate(const cTimer* t) = 0;
};

class cTimerDateVar : public cTimerVar {
 public:
    cTimerDateVar() : cTimerVar("%timer.date%") {}
    string Evaluate(const cTimer* t) 
	{ 
	    if (!t) return "";
	    return DATESTRING(t->StartTime());
	}
};

class cTimerStartVar : public cTimerVar {
 public:
    cTimerStartVar() : cTimerVar("%timer.start%") {}
    string Evaluate(const cTimer* t) 
	{ 
	    if (!t) return "";
	    return TIMESTRING(t->StartTime());
	}
};

class cTimerStopVar : public cTimerVar {
 public:
    cTimerStopVar() : cTimerVar("%timer.stop%") {}
    string Evaluate(const cTimer* t) 
	{ 
	    if (!t) return "";
	    return TIMESTRING(t->StopTime());
	}
};

class cTimerFileVar : public cTimerVar {
 public:
    cTimerFileVar() : cTimerVar("%timer.file%") {}
    string Evaluate(const cTimer* t) 
	{ 
	    if (!t) return "";
	    return t->File();
	}
};

class cTimerChnrVar : public cTimerVar {
 public:
    cTimerChnrVar() : cTimerVar("%timer.chnr%") {}
    string Evaluate(const cTimer* t) 
	{ 
	    if (!t || !t->Channel()) return "";
	    return IntToString(t->Channel()->Number());
	}
};

class cTimerChannelShortVar : public cTimerVar {
public:
    cTimerChannelShortVar() : cTimerVar("%timer.chsh%") {}
    string Evaluate(const cTimer* t) 
	{ 
	    if (!t || !t->Channel()) return "";
	    return t->Channel()->ShortName(true);
	}
};

class cTimerChannelLongVar : public cTimerVar {
public:
    cTimerChannelLongVar() : cTimerVar("%timer.chlng%") {}
    string Evaluate(const cTimer* t) 
	{ 
	    if (!t || !t->Channel()) return "";
	    return t->Channel()->Name();
	}
};

class cTimerSearchVar : public cTimerVar {
 public:
    cTimerSearchVar() : cTimerVar("%timer.search%") {}
    string Evaluate(const cTimer* t) 
	{ 
	    if (!t) return "";
	    cSearchExt* s = TriggeredFromSearchTimer(t);
	    if (!s) return "";
	    return s->search;
	}
};

class cTimerSearchIDVar : public cTimerVar {
 public:
    cTimerSearchIDVar() : cTimerVar("%timer.searchid%") {}
    string Evaluate(const cTimer* t) 
	{ 
	    if (!t) return "";
	    int ID = TriggeredFromSearchTimerID(t);
	    if (ID < 0) return "";
	    return IntToString(ID);
	}
};

class cUserVars : public cList<cUserVar> {
 public:
    cTitleVar titleVar;
    cSubtitleVar subtitleVar;
    cSummaryVar summaryVar;
    cHTMLSummaryVar htmlsummaryVar;
    cEventIDVar eventIDVar;
    cTimeVar timeVar;
    cTimeEndVar timeEndVar;
    cTime_wVar time_wVar;
    cTime_dVar time_dVar;
    cTime_lngVar time_lngVar;
    cDateVar dateVar;
    cDateShortVar dateShortVar;
    cChannelNrVar chnrVar;
    cChannelShortVar chShortVar;
    cChannelLongVar chLongVar;
    cChannelDataVar chDataVar;

    cColonVar colonVar;
    cDateNowVar dateNowVar;
    cDateShortNowVar dateShortNowVar;
    cTimeNowVar timeNowVar;

    cTimerDateVar timerDateVar;
    cTimerStartVar timerStartVar;
    cTimerStopVar timerStopVar;
    cTimerFileVar timerFileVar;
    cTimerChnrVar timerChnrVar;
    cTimerChannelShortVar timerChShortVar;
    cTimerChannelLongVar timerChLongVar;
    cTimerSearchVar timerSearchVar; 
    cTimerSearchIDVar timerSearchIDVar; 

    map<string, cExtEPGVar*> extEPGVars;
    set<cUserVar*> userVars;
    map<string, cInternalVar*> internalVars;
    map<string, cTimerVar*> internalTimerVars;

    void InitInternalVars()
	{
	    internalVars[titleVar.Name()] = &titleVar;
	    internalVars[subtitleVar.Name()] = &subtitleVar;
	    internalVars[summaryVar.Name()] = &summaryVar;
	    internalVars[htmlsummaryVar.Name()] = &htmlsummaryVar;
	    internalVars[eventIDVar.Name()] = &eventIDVar;
	    internalVars[timeVar.Name()] = &timeVar;
	    internalVars[timeEndVar.Name()] = &timeEndVar;
	    internalVars[time_wVar.Name()] = &time_wVar;
	    internalVars[time_dVar.Name()] =  &time_dVar;
	    internalVars[time_lngVar.Name()] = &time_lngVar;
	    internalVars[dateVar.Name()] = &dateVar;
	    internalVars[dateShortVar.Name()] = &dateShortVar;
	    internalVars[chnrVar.Name()] = &chnrVar;
	    internalVars[chShortVar.Name()] = &chShortVar;
	    internalVars[chLongVar.Name()] = &chLongVar;
	    internalVars[chDataVar.Name()] = &chDataVar;

	    internalVars[colonVar.Name()] = &colonVar;
	    internalVars[dateNowVar.Name()] = &dateNowVar;
	    internalVars[dateShortNowVar.Name()] = &dateShortNowVar;
	    internalVars[timeNowVar.Name()] = &timeNowVar;

	    internalTimerVars[timerDateVar.Name()] = &timerDateVar;	    
	    internalTimerVars[timerStartVar.Name()] = &timerStartVar;
	    internalTimerVars[timerStopVar.Name()] = &timerStopVar;
	    internalTimerVars[timerFileVar.Name()] = &timerFileVar;
	    internalTimerVars[timerChnrVar.Name()] = &timerChnrVar;
	    internalTimerVars[timerChShortVar.Name()] = &timerChShortVar;
	    internalTimerVars[timerChLongVar.Name()] = &timerChLongVar;
	    internalTimerVars[timerSearchVar.Name()] = &timerSearchVar;
	    internalTimerVars[timerSearchIDVar.Name()] = &timerSearchIDVar;
	}

    void InitExtEPGVars()
	{
	    cSearchExtCat* SearchExtCat = SearchExtCats.First();
	    while (SearchExtCat) 
	    {
		string varName = string("%") + SearchExtCat->name + string("%");
		std::transform(varName.begin(), varName.end(), varName.begin(), tolower);
		cExtEPGVar* extEPGVar = new cExtEPGVar(varName);
		extEPGVars[varName] =  extEPGVar;
		SearchExtCat = SearchExtCats.Next(SearchExtCat);
	    }
	}
    void ResetCache()
	{
	    cUserVar* var = First();
	    while (var) 
	    {
		var->ResetCache();
		var = Next(var);
	    }
	}
    ~cUserVars()
	{
	    std::map<string, cExtEPGVar*>::iterator evar;
	    for (evar = extEPGVars.begin(); evar != extEPGVars.end(); evar++)
		delete evar->second;
	    extEPGVars.clear();

	    std::set<cUserVar*>::iterator uvar;
	    for (uvar = userVars.begin(); uvar != userVars.end(); uvar++)
		delete (*uvar);
	    userVars.clear();
	}
    cUserVar* GetFromName(const string& varName);
};

extern cUserVars UserVars;

class cUserVarLine : public cListObject {
public:
  bool Parse(char *s);
};


class cUserVarFile : public cConfig<cUserVarLine> 
{
public:
    cUserVarFile() { UserVars.Clear(); };
};

class cVarExpr 
{
    string expr;
 public:
    set<cUserVar*> usedVars;
    cVarExpr(const string& Expr) : expr(Expr) {} 
    string Evaluate(const cEvent* e = NULL);
    string Evaluate(const cTimer* t);
    bool DependsOnVar(const string& varName, const cEvent* e);
};

#endif
